# file transfer management class
# coding: utf-8

##   Copyright (C) 2010 Coopi <gnu@coopi.org>
## 
##   This program is free software; you can redistribute it and/or
##   modify it under the terms of the GNU General Public License as
##   published by the Free Software Foundation; either version 2 of the
##   License, or (at your option) any later version.
## 
##   This program is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##   General Public License for more details.
## 
##   You should have received a copy of the GNU General Public
##   License along with this program; if not, write to the
##   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import socket
import gobject
import sys
import random
import os
import pygtk
pygtk.require('2.0')
import gtk

class file_transfer:

	# cancel a file transfer (both send and receive)
	def cancel_transfer(self):
		# close file and socket
		self.f.close()
		self.conn.close()

		# close also the primary socket if it's a receiving operation
		if self.transfer_type == 0: self.socket.close()

		if self.pygeon.debug: print _("File transfer has been cancelled")
				
		# delete the box into file transfer window
		if self.pygeon.debug: print _("I delete the box in the file transfer window. Now the file transfers are: ") + str(len(self.pygeon.file_manager_class.elenco.get_children())-1)
		self.box.destroy()
		if len(self.pygeon.file_manager_class.elenco.get_children()) == 0: self.pygeon.file_manager_class.cmdClose.clicked()
				
		# delete the file if is a file receiving
		if self.transfer_type == 0 and os.path.isfile(os.path.normpath(self.pygeon.files_path + "/" + self.filename)):
			os.remove(os.path.normpath(self.pygeon.files_path + "/" + self.filename))
			if self.pygeon.debug: print _("I delete the file ") + self.pygeon.files_path + "/" + self.filename
				
		# remove the class from the list
		self.pygeon.file_transfers.remove(self)
		if self.pygeon.debug: print _("I delete the class from the list. Now the list size is: ") + str(len(self.pygeon.file_transfers))
				
		# notification and warning message
		if self.transfer_type == 1: self.filename = os.path.basename(self.filename)
		self.pygeon.show_notification("file-transfer", "", _("File transfer cancelled"), _("The transfer of ") + self.filename + _(" has been cancelled"))
		self.pygeon.message_box(_("The transfer of ") + self.filename + _(" has been cancelled"), _("File transfer cancelled"), gtk.MESSAGE_WARNING, gtk.BUTTONS_OK, self.pygeon.winMain)

	# close a transfer (both send and receive)
	def close_transfer(self):
		# close file and socket
		self.f.close()
		self.conn.close()
		
		# close also the primary socket if it's a receiving operation
		if self.transfer_type == 0: self.socket.close()

		if self.pygeon.debug: print _("File transfer is finished")
					
		# delete the box into file transfer window
		if self.pygeon.debug: print _("I delete the box in the file transfer window. Now the file transfers are: ") + str(len(self.pygeon.file_manager_class.elenco.get_children())-1)
		self.box.destroy()
		if len(self.pygeon.file_manager_class.elenco.get_children()) == 0: self.pygeon.file_manager_class.cmdClose.clicked()
						
		# remove the class from the list
		self.pygeon.file_transfers.remove(self)
		if self.pygeon.debug: print _("I delete the class from the list. Now the list size is: ") + str(len(self.pygeon.file_transfers))

		# check the file transfer status if is a receiving action
		# send a file transfer finished message if is a sending action
		if self.transfer_type == 0:
			if self.pygeon.debug: print _("I check the file size")

			if self.check_file():
				# file transfer is ok
				if self.pygeon.debug: print _("File transfer was successful")

				self.pygeon.send_file_transfer_ok(self.mittente, self.fid, self.filename)

				# notification
				self.pygeon.show_notification("file-transfer", self.pygeon.get_name_from_username(self.mittente.split('/')[0]), _("File transfer finished"), _("The file ") + self.filename + _(" sent from ") + self.pygeon.get_name_from_username(self.mittente.split('/')[0]) + _(" was received correctly"))
			else:
				# the file is corrupted
				if self.pygeon.debug: print _("File transfer is failed. The file is corrupted")

				self.pygeon.send_file_transfer_error(self.mittente, self.fid, self.filename)
				
				# notification and error message
				self.pygeon.show_notification("file-transfer", self.pygeon.get_name_from_username(self.mittente.split('/')[0]), _("File transfer failed"), _("The sending of ") + self.filename + _(" from ") + self.pygeon.get_name_from_username(self.mittente.split('/')[0]) + _(" is failed"))
				self.pygeon.message_box(_("Transfer of ") + self.filename + _(" from ") + self.pygeon.get_name_from_username(self.mittente.split('/')[0]) + _(" is failed"), _("File transfer failed"), gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, self.pygeon.winMain)

		else: self.pygeon.send_file_transfer_finished(to=self.destinatario, fid=self.fid)

	# manage the timeout
	def manage_timeout(self):
		if self.pygeon.ddebug: print _("Timeout")

		try:
			self.timeout += self.speed
		except:
			self.timeout = self.speed

		if self.timeout >= self.max_timeout: self.cancelled = True

	# wait a connection from the sender
	def start_accept(self):
		if self.pygeon.debug: print _("I'm waiting for accept")

		try:
			# start receiving file if the client is connected
			self.conn, addr = self.socket.accept()
			self.conn.setblocking(0)
			
			if self.pygeon.debug: print _("The sender is connected")

			gobject.timeout_add(self.speed, self.start_receive)

			return False # stop timer
		except:
			return True

	# wait to receive data from the sender
	def start_receive(self):
		if not self.receiving:
			self.receiving = True

			# check if the file transfer has been cancelled
			if self.cancelled:
				self.cancel_transfer()

				self.receiving = False
				return False # stop timer
			else:
				try:
					data = self.conn.recv(self.chunk)

					if data is None or data == "":
						# check if file transfer is finished or if there is timeout
						if self.finished and self.tot_received == self.size: self.close_transfer()
						else: self.manage_timeout()

						self.receiving = False
						return not (self.finished and self.tot_received == self.size)
					else:
						self.timeout = 0 # reset timeout

						if self.tot_received == 0:
							# first receiving time, create the file
							self.f = open(os.path.normpath(self.pygeon.files_path + "/" + self.filename), "wb")

							if self.pygeon.debug: print _("File transfer started")

						# write data to file
						self.f.write(data)
						self.tot_received += len(data)

						# set the progress bar
						self.current_step = int((self.tot_received * self.steps) / self.size)
						self.pygeon.set_progress_transfer(self.progress, self.current_step, self.tot_received, self.size, True)

						if self.pygeon.ddebug: print _("Bytes received: ") + str(self.tot_received)

						self.receiving = False
						return True
				except:
					# check if file transfer is finished or if there is timeout
					if self.finished and self.tot_received == self.size: self.close_transfer()
					else: self.manage_timeout()

					self.receiving = False
					return not (self.finished and self.tot_received == self.size)

	# file sending
	def start_send(self):
		if not self.sending:
			self.sending = True

			# check if file transfer has been cancelled
			if self.cancelled:
				self.cancel_transfer()

				self.sending = False
				return False # stop timer
			else:
				if self.tot_sent == 0:
					# if nothing has been sent open the file in read mode
					self.f = open(self.filename, "rb")

					if self.pygeon.debug: print _("I open the file in read mode")

				br = self.f.read(self.chunk) # read the file
				if len(br) > 0: # if the operation isn't finished

					try:
						b = self.conn.send(br)#, self.chunk)
					except socket.error, e:
						#print "Errore: %s" % e
						#sys.exit()
						b = 0

					self.tot_sent += b

					# go to the right position into the file if it read more than it sent
					if b != br: self.f.seek(self.tot_sent)

					if b > 0: # send without problems
						self.timeout = 0 # reset timeout

						self.current_step = int((self.tot_sent * self.steps) / self.size)
						self.pygeon.set_progress_transfer(self.progress, self.current_step, self.tot_sent, self.size, True)

						if self.pygeon.ddebug: print _("I sent ") + str(self.tot_sent)

						self.sending = False
						return True
					else: # read the file but it cannot send
						self.manage_timeout()

						self.sending = False
						return True
				else: # file reading is finished, close the file transfer
					self.close_transfer()

					self.sending = False
					return False # stop timer

	# wait for receipient response (accept or refuse)
	def wait_accept(self):
		if self.accepted is None: # no response
			if self.pygeon.ddebug: print _("I'm waiting for a response")

			return True
		elif self.accepted: # file transfer accepted
			if self.pygeon.debug: print _("The request has been accepted. I proceed with the file transfer")

			# create an entry in file_manager class and initialize the objects
			info = self.pygeon.get_name_from_username(self.destinatario.split('/')[0]) + " - " + os.path.basename(self.filename)
			if len(info) > 40: info = info[:40] + "..."

			lblInfo, imgTransfer, self.progress, self.box = self.pygeon.file_manager_class.add_file_transfer(self.fid, self.destinatario)
			lblInfo.set_markup("<b>" + info + "</b>")
			imgTransfer.set_from_file()

			self.pygeon.create_progress(self.progress, self.steps, True)
			self.current_step = 0

			self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self.conn.connect((self.host, self.port))
			self.conn.setblocking(0)

			# if the receipient has accepted start the socket
			gobject.timeout_add(self.speed, self.start_send)

			return False # stop timer
		elif not self.accepted: # file transfer has been refused
			# remove the class from the list
			self.pygeon.file_transfers.remove(self)
			if self.pygeon.debug: print _("I delete the class from the list. Now the list size is: ") + str(len(self.pygeon.file_transfers))

			# notification
			self.pygeon.show_notification("file-transfer", self.pygeon.get_name_from_username(self.destinatario.split('/')[0]), _("File transfer refused"), _("Transfer of ") + os.path.basename(self.filename) + _(" has been refused from ") + self.pygeon.get_name_from_username(self.destinatario.split('/')[0]))

			# show also an error window
			self.pygeon.message_box(_("Transfer of ") + os.path.basename(self.filename) + _(" has been refused from ") + self.pygeon.get_name_from_username(self.destinatario.split('/')[0]), _("File transfer refused"), gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, self.pygeon.winMain)

			return False # stop timer

	# returns a random port from 50000 to 60000
	def get_random_port(self):
		return random.randint(50000, 60000)

	# return true if the port is in used, otherwise return false
	def port_in_use(self, p):
		for c in self.pygeon.file_transfers:
			if c.port == p: return True

		return False

	# check if the file has been received correctly (size)
	def check_file(self):
		return self.size == os.path.getsize(os.path.normpath(self.pygeon.files_path + "/" + self.filename))

	# constructor
	def __init__(self, pygeon, filename=None, mittente=None, host=None, port=None, destinatario=None, fid=None, size=None, hash=None):
		self.pygeon = pygeon # main class

		self.f = None # file
		self.fid = fid # file transfer id
		self.filename = filename # filename
		self.cancelled = False # file transfer cancelled
		self.chunk = 102400 # bytes sent/received in a single operation
		self.speed = 15 # timer timeout in milliseconds
		self.max_timeout = 60000 # max timeout time

		if not mittente is None:
			# receiving a file
			self.receiving = False # semaphore
			self.tot_received = 0 # bytes received
			self.finished = False # file transfer finished
			self.mittente = mittente # file sender
			self.size = size # file size
			self.hash = hash # file hash
			self.port = 0 # port value
			self.steps = int(self.size / self.chunk) # number of steps for progress bar
			self.transfer_type = 0 # file transfer type => receive

			while 1:
				# get a random port for socket (from 50000 to 60000)
				while 1:
					p = self.get_random_port()

					# check if the port is already in use
					if not self.port_in_use(p): break
				self.port = p

				self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
				self.socket.setblocking(0)



				try:
					self.socket.bind((self.pygeon.get_local_ip(), self.port))
					break
				except:
					pass
				
			self.socket.listen(1)

			# create an entry into the file_manager class and set the objects
			info = self.pygeon.get_name_from_username(self.mittente.split('/')[0]) + " - " + self.filename
			if len(info) > 40: info = info[:40] + "..."

			lblInfo, imgTransfer, self.progress, self.box = self.pygeon.file_manager_class.add_file_transfer(self.fid, self.mittente)
			lblInfo.set_markup("<b>" + info + "</b>")
			imgTransfer.set_from_file(os.path.normpath(self.pygeon.icon_theme_application_path + "download.png"))
			
			self.pygeon.create_progress(self.progress, self.steps, True)
			self.current_step = 0
			
			gobject.timeout_add(500, self.start_accept)

			# tell the sender to send the file
			self.pygeon.send_file_transfer_accepted(to=mittente, host=self.pygeon.get_local_ip(), port=self.port, fid=self.fid)
		elif not destinatario is None:
			self.sending = False # semaphore
			self.tot_sent = 0 # bytes sent
			self.size = os.path.getsize(os.path.normpath(self.filename)) # file size
			self.steps = int(self.size / self.chunk) # number of steps for progress bar
			self.destinatario = destinatario # file receipient
			self.host = None # host to connect (initially null)
			self.port = None # port (initially null)
			self.accepted = None # receipient acceptance (initially null)
			self.transfer_type = 1 # file transfer type => send

			# start timer
			gobject.timeout_add(500, self.wait_accept)
